var bcrypt = require("bcrypt");
var mc = require('mongodb').MongoClient;
var playersCollection;
var roomsCollection;

var defaultRoom = {   name: "theVoid",
                      title: "The Void",
                      description: "You are in the Void.  How did you get here?  There are no exits.",
                      roomExits: ['theVoid']
                  };        

var connectToDBs = function(callback) {
    mc.connect('mongodb://localhost/adventure-demo', function(err, db) {
        if (err) {
            throw err;
        }
        
        playersCollection = db.collection('players');
        roomsCollection = db.collection('rooms');

        if (callback) {
            callback();
        }
    });
}

var savePlayer = function(player) {
    playersCollection.update({"playername": player.playername},
                             player, function(err, count) {
                                 if (err) {
                                     console.log(
                                         "Couldn't save player state");
                                 }
                             });
}


var index = function(req, res) {
    if (req.session.player) {
        res.redirect("/game");
    } else {
        res.render('index', { title: 'COMP 2406 AJAX Adventure Demo', 
                              error: req.query.error });
    }
}

var register = function(req, res) {
    var playername = req.body.playeredname;
    var password = req.body.password;

    var addPlayer = function(err, players) {
        if(players.length!=0){
            res.redirect("/?error=name already exists");        
            return;
        }
        
        //generate a salt, with 10 rounds (2^10 iterations)
        bcrypt.genSalt(10, function(err, salt) {
            bcrypt.hash(password, salt, function(err, hash) {
                var newPlayer = {
                    playername: playername,
                    password: hash,
                    room: "bridge"
                };
                
                playersCollection.insert(newPlayer, function(err, newPlayers){
                    if (err) {
                        throw err;
                    } else {
                        res.render('registered', 
                                   { playername: newPlayers[0].playername });
                    }
                });    
            });
        });     
    };  
    
    playersCollection.find({playername: playername}).toArray(addPlayer);
}

var checkPass = function(req, res, name, password, sessionName,
                         successCallback) {
    playersCollection.findOne({playername: name}, function(err, name){
        
        if (err || !name){
            req.session.destroy(function(err) {
                res.redirect("/?error=invalid name or password");       
            });
            return;
        }
        
        bcrypt.compare(password, name.password, function(err, authenticated){
            if (authenticated) {
                req.session[sessionName] = name;
                delete req.session[sessionName]._id;
                successCallback(req, res);
            } else {
                req.session.destroy(function(err) {
                    res.redirect("/?error=invalid name or password");
                });
            }
        });
    });
}

var start = function(req, res) {

    var successCallback = function(req, res) {
        req.session.player.online = true;
        savePlayer(req.session.player);  // finishes asynchronously
        res.redirect("/game");
    }

    var playername = req.body.playeredname;
    var password = req.body.password;
    
    checkPass(req, res, playername, password, "player", successCallback);
}

var quit = function(req, res){

    req.session.player.online = false; // assuming only logged in once
    // savePlayer here is safe because req.session.destroy won't delete 
    // the player object; instead it will persist as long as needed
    // for savePlayer to finish (referenced by the local player variable
    // in savePlayer).
    savePlayer(req.session.player);
    
    req.session.destroy(function(err){
        if (err) {
            console.log("Error: %s", err);
        }
        res.redirect("/");
    }); 
}

var game = function(req, res) {
    if (req.session.player) {
        res.render("room.jade", {title: "AJAX Adventure Demo"});
    } else {
        res.redirect("/");
    }
}

var doAction = function(req, res) {
    var action = req.body.action;
    var room = req.body.room;

    if (!req.session.player) {
        res.send("Error: NotLoggedIn");
        return;
    }

    if (action === "move") {
        req.session.player.room = room;
        savePlayer(req.session.player);
        res.send("Success");
    } else {
        res.send("Error: InvalidAction");
    }
}


var getContents = function(req, res) {

    var roomObj, theRoom;

    var returnRoom = function(err, presentPlayers) {
        var players = "";

        if (err) {
            console.log("Couldn't get players in current room.");
            presentPlayers = [];
        }

        presentPlayers.forEach(function (p) {
            players = players + p.playername + " ";
        });
        
        if (players !== "") {
            players = "Players present: " + players;
        }

        roomObj.players = players;
        res.send(roomObj);
    }

    var retrievedRoom = function(err, room) {
        if (err || !room) {
            console.log("Couldn't find room " + theRoom);
            room = defaultRoom;
        }
        
        roomObj = room;

        if (req.session.player) {
            playersCollection.find({online: true,
                                    room: theRoom,
                                    playername:
                                    {$ne: req.session.player.playername}},
                                   {"_id": false, "playername": true}
                                  ).toArray(returnRoom);
        } else {
            returnRoom(false, []);
        }
    }

    if (req.session.player) {
        theRoom = req.session.player.room;
    } else if (req.session.editorName) {
        theRoom = req.query.theRoom;
    } else {
        console.log("got invalid getContents request");
        res.send(undefined);
    }
    
    roomsCollection.findOne({name: theRoom}, retrievedRoom);
}

var startEditor = function(req, res) {
    var editorName = req.body.playeredname;
    var password = req.body.password;

    var successCallback = function(req, res) {
        res.redirect("/editor");
    }
    
    checkPass(req, res, editorName, password, "editorName", successCallback);
}

var getActiveRooms = function(callback) {

    var returnRooms = function(err, room) {
        if (err || !room) {
            console.log("Couldn't find roomList");
            room = {activeRooms: ["theVoid"]};
        }
        callback(room.activeRooms);
    }

    roomsCollection.findOne({name: "roomList"}, returnRooms);
}

var editor = function(req, res) {
    var renderEditor = function(activeRooms) {
        activeRooms.unshift("");
        res.render("editor.jade", {editorName:
                                   req.session.editorName.playername,
                                   rooms: activeRooms});
    }

    if (req.session.editorName) {
        getActiveRooms(renderEditor);
    } else {
        res.redirect("/");
    }
}


var addRoomToActive = function(roomName) {
    var checkErrors = function(err, count) {
        if (err | count != 1) {
            console.log("Problem adding room to activeList" + roomName);
        }
    }

    var callback = function(activeRooms) {
        if (activeRooms.indexOf(roomName) === -1) {
            activeRooms.push(roomName);
            roomsCollection.update({name: "roomList"},
                                   {name: "roomList",
                                    activeRooms: activeRooms},
                                   checkErrors);
        } // otherwise room is already in the list
    }

    getActiveRooms(callback);
}

var updateRoom = function(req, res) {
    
    var updateResponse = function(err, count) {
        if (err) {
            res.render("editorSave", {title: "Room Save Status",
                                      success: false});
        } else {
            res.render("editorSave", {title: "Room Save Status", 
                                      success: true});
        }
    }

    var newRoom = {
        name: req.body.roomName,
        title: req.body.roomTitle,
        description: req.body.description,
        activity: req.body.activity,
    };

    addRoomToActive(newRoom.name);

    if (req.body.exits) {
        newRoom.roomExits = req.body.exits.split(",")
    } else {
        newRoom.roomExits = [];
    }

    roomsCollection.update({"name": newRoom.name},
                           newRoom,
                           {upsert: true},
                           updateResponse);
}

exports.index = index;
exports.register = register;
exports.start = start;
exports.quit = quit;
exports.game = game;
exports.connectToDBs = connectToDBs;
exports.doAction = doAction;
exports.getContents = getContents;
exports.startEditor = startEditor;
exports.editor = editor;
exports.updateRoom = updateRoom;
